use std::{
    cell::{Cell, RefCell},
    collections::HashMap,
    rc::Rc,
};

use mio::{Events, Interest, Poll, Token};

use crate::{prelude::ChannelSpec, service::event::NoPriorityTag, Fleck};

pub mod builder;
// pub(super) mod stream;
// pub(super) mod udp;

use builder::DefaultServiceBuilder;

lazy_static::lazy_static! {
    pub static ref MULTICAST_ADDRESS : std::net::SocketAddrV4 =
        std::net::SocketAddrV4::new(
            std::net::Ipv4Addr::new(239, 0, 2, 35),
            3535
        );
}

pub const DEFAULT_PORT: u16 = 3535;

pub(super) trait AbstractFdService: 'static {
    fn ready_read(&self, api: &Fleck);
    fn ready_write(&self, api: &Fleck);
    fn token(&self) -> Token;
}

pub(super) struct IO {
    poll: Rc<RefCell<mio::Poll>>,
    services: RefCell<HashMap<Token, Rc<dyn AbstractFdService>>>,
}

impl IO {
    pub(super) fn new() -> Self {
        Self {
            poll: Rc::new(RefCell::new(
                mio::Poll::new().expect("couldn't create mio::Poll"),
            )),
            services: Default::default(),
        }
    }

    pub(super) fn run(&self, api: &Fleck) {
        let mut events = mio::Events::with_capacity(128);
        loop {
            log::trace!("Waiting for events...");
            let pr = self.poll.borrow_mut().poll(&mut events, None);
            log::trace!("Polled!");
            if pr.is_err() {
                continue;
            }
            log::trace!("Waking up for events");
            let svcs = self.services.borrow();
            for evt in &events {
                let svc = svcs.get(&evt.token()).unwrap();
                if evt.is_readable() {
                    svc.ready_read(api);
                }
                if evt.is_writable() {
                    svc.ready_write(api);
                }
            }

            api.services().event_root().fire_all(api);
        }
    }

    pub(super) fn add_service<'l>(&self, fleck: &'l Fleck) -> DefaultServiceBuilder<'l> {
        let tok = Token(self.services.borrow().len() + 1);

        builder::DefaultServiceBuilder::new(fleck, self.poll.clone(), tok)
    }

    /*pub(super) fn add_service<S: AbstractFdService>(&mut self, srv: S) {
        let tok = Token(self.services.len() + 1);
        self.services.insert(tok, Rc::new(srv));
    }

    pub(super) fn add_udp_service<CB: 'static + udp::UdpCallback>(&mut self, udp: mio::net::UdpSocket, on_read: CB) -> impl Fn(std::net::SocketAddr, Vec<u8>) {
        let token = Token(self.services.len() + 1);
        let srv = Rc::new(udp::UdpService {
            socket: RefCell::new(udp),
            token,
            poll: self.poll.clone(),
            send_queue: Default::default(),
            write_registered: Cell::new(false),
            read_callback: Rc::new(on_read),
        });
        self.services.insert(token, srv);

        |addr, data| {

        }
        /*self.services.insert(token, Box::new(udp::UdpService::<Incoming, Outgoing, DataType> {
            socket: RefCell::new(udp),
            token,
            poll: self.poll.clone(),
            send_queue: Default::default(),
            write_registered: std::cell::Cell::new(false),
            _ghost: std::marker::PhantomData
        }));*/
    }*/
}

/*pub struct FleckIO {
    poll: Rc<RefCell<mio::Poll>>,
    services: HashMap<Token, Box<dyn AbstractFdServiceHelper>>,
    // local: Vec<std::rc::Rc<dyn IOChannel>>,
    // global: std::rc::Rc<SocketWrapperChannel>,

    // minor: timerfd::TimerFd,
    // major: timerfd::TimerFd,
}*/

/*
struct SocketWrapperChannel {
    poll: Rc<RefCell<mio::Poll>>,
    socket: RefCell<mio::net::UdpSocket>,
    queue: RefCell<VecDeque<(Vec<u8>, std::net::SocketAddr)>>,
    write_registered: Cell<bool>,
}

impl SocketWrapperChannel {
    fn new(poll: Rc<RefCell<mio::Poll>>, socket: mio::net::UdpSocket) -> Self {
        Self {
            poll,
            socket: RefCell::new(socket),
            queue: Default::default(),
            write_registered: Cell::new(false),
        }
    }

    fn read_packets(&self, feedback: &dyn IOFeedback) {
    }

    fn write_packets(&self) {
        let mut queue = self.queue.borrow_mut();
        while !queue.is_empty() {
            let pkt = queue.front().unwrap();
            match self.socket.borrow().send_to(&pkt.0, pkt.1) {
                Ok(len) => {
                    if len != pkt.0.len() {
                        log::warn!(
                            "Packet of {} bytes truncated to {} bytes when sending",
                            pkt.0.len(),
                            len
                        );
                    }
                    queue.pop_front();
                }
                Err(e) => {
                    log::info!("Error while sending packet: {:?}", e);
                    break;
                }
            }
        }

        if queue.is_empty() {
            self.poll
                .borrow()
                .registry()
                .reregister(
                    self.socket.borrow_mut().deref_mut(),
                    FleckIO::GLOBAL,
                    Interest::READABLE,
                )
                .expect("couldn't update interest?");
            self.write_registered.set(false);
        }
    }
}
*/

/*impl IOChannel for SocketWrapperChannel {
    fn send_packet(&self, packet: &mut Packet) {
        assert!(packet.data.is_some());
        assert!(packet.addr.is_some());

        self.queue
            .borrow_mut()
            .push_back((packet.data.clone().unwrap(), packet.addr.unwrap()));
        if !self.write_registered.get() {
            self.write_registered.set(true);
            self.poll
                .borrow()
                .registry()
                .reregister(
                    self.socket.borrow_mut().deref_mut(),
                    FleckIO::GLOBAL,
                    Interest::READABLE.add(Interest::WRITABLE),
                )
                .expect("couldn't update interest?");
        }
    }
}*/

/*impl Default for FleckIO {
    fn default() -> Self {
        let poll = Poll::new().expect("couldn't create mio::Poll");

        let poll = Rc::new(RefCell::new(poll));

        Self {
            poll,
            ..Default::default()
        }
    }
}
*/

/*
impl FleckIO {
    fn handle_event(&self, api: &Fleck, event: &mio::event::Event) {
        /*match event.token() {
            Self::GLOBAL => {
                if event.is_readable() {
                    self.global.read_packets(feedback);
                }
                if event.is_writable() {
                    self.global.write_packets();
                }
            }
            Self::MINOR_TICK => {
                assert!(event.is_readable());
                self.minor.read();
                feedback.minor_tick();
            }
            Self::MAJOR_TICK => {
                assert!(event.is_readable());
                self.major.read();
                feedback.major_tick();
            }
            _ => {
                println!("unknown token!");
            }
        }*/
    }

    pub(crate) fn poll(&self, api: &Fleck) {
        let mut events = Events::with_capacity(128);
        loop {
            let pr = self.poll.borrow_mut().poll(&mut events, None);
            match pr {
                Ok(()) => {
                    for evt in &events {
                        self.handle_event(api, evt);
                    }
                }
                Err(_e) => {}
            }
        }
    }

    /*pub(crate) fn local(&self) -> std::rc::Rc<dyn IOChannel> {
        // HACK: for now we just use the global
        self.global.clone()
    }

    pub(crate) fn global(&self) -> std::rc::Rc<dyn IOChannel> {
        self.global.clone()
    }*/

    const GLOBAL: Token = Token(0);
    const MAJOR_TICK: Token = Token(1);
    const MINOR_TICK: Token = Token(2);
}
*/

/*

       // minor tick registration
       let mut minor = timerfd::TimerFd::new().unwrap();
       minor.set_state(
           timerfd::TimerState::Periodic {
               current: std::time::Duration::new(1, 0),
               interval: std::time::Duration::new(1, 0),
           },
           timerfd::SetTimeFlags::Default,
       );
       poll.registry()
           .register(
               &mut SourceFd(&minor.as_raw_fd()),
               Self::MINOR_TICK,
               Interest::READABLE,
           )
           .expect("couldn't register read interest for minor timer?");

       // major tick registration
       let mut major = timerfd::TimerFd::new().unwrap();
       major.set_state(
           timerfd::TimerState::Periodic {
               current: std::time::Duration::new(1, 0),
               interval: std::time::Duration::new(15, 0),
           },
           timerfd::SetTimeFlags::Default,
       );
       poll.registry()
           .register(
               &mut SourceFd(&major.as_raw_fd()),
               Self::MAJOR_TICK,
               Interest::READABLE,
           )
           .expect("couldn't register read interest for major timer?");

       // global socket
       let mut global_socket = mio::net::UdpSocket::bind(
           std::net::SocketAddrV4::new(std::net::Ipv4Addr::UNSPECIFIED, DEFAULT_PORT).into(),
       )
       .expect("couldn't listen on UDP port?");
       global_socket
           .join_multicast_v4(&MULTICAST_ADDRESS.ip(), &std::net::Ipv4Addr::UNSPECIFIED)
           .expect("couldn't join multicast group?");

       poll.registry()
           .register(&mut global_socket, Self::GLOBAL, Interest::READABLE)
           .expect("couldn't register read interest in UDP port?");
*
* */
