use std::cell::Cell;
use std::cell::RefCell;
use std::collections::VecDeque;
use std::fs::File;
use std::io::Read;
use std::io::Write;
use std::os::unix::prelude::AsRawFd;
use std::rc::Rc;

use mio::unix::SourceFd;
use mio::Interest;
use mio::Token;

use super::AbstractFdService;
use crate::prelude::*;
use crate::service::event::NoPriorityTag;
use crate::Fleck;

pub trait StreamData: 'static + Into<Vec<u8>> + From<Vec<u8>> {}
pub struct StreamService<
    Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>,
    Outgoing: ChannelSpec<Data = DataType>,
    DataType: StreamData,
> {
    stream: RefCell<File>,
    token: Token,
    poll: Rc<RefCell<mio::Poll>>,
    send_queue: RefCell<VecDeque<u8>>,
    write_registered: Cell<bool>,
    _ghost: std::marker::PhantomData<(Incoming, Outgoing, DataType)>,
}

impl<
        Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>,
        Outgoing: ChannelSpec<Priority = NoPriorityTag, Data = DataType>,
        DataType: StreamData,
    > Service for Rc<StreamService<Incoming, Outgoing, DataType>>
{
    fn register_channels(&self, eroot: &mut crate::prelude::EventRoot) {
        eroot
            .channel::<Outgoing>()
            .sub_opt(self, StreamService::handle_outgoing);
    }
}

impl<
        Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>,
        Outgoing: ChannelSpec<Priority = NoPriorityTag, Data = DataType>,
        DataType: StreamData,
    > StreamService<Incoming, Outgoing, DataType>
{
    fn handle_outgoing(&self, _api: &Fleck, packet: DataType) -> Option<DataType> {
        self.send_queue.borrow_mut().extend(packet.into().iter());

        self.register_write_interest();

        None
    }

    fn clear_write_interest(&self) {
        self.poll
            .borrow()
            .registry()
            .reregister(
                &mut SourceFd(&self.stream.borrow().as_raw_fd()),
                self.token,
                Interest::READABLE,
            )
            .expect("couldn't update interest?");
        self.write_registered.set(false);
    }

    fn register_write_interest(&self) {
        if self.write_registered.get() {
            return;
        }
        self.poll
            .borrow()
            .registry()
            .reregister(
                &mut SourceFd(&self.stream.borrow().as_raw_fd()),
                self.token,
                Interest::READABLE.add(Interest::WRITABLE),
            )
            .expect("couldn't update interest?");
        self.write_registered.set(true);
    }
}

impl<
        Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>,
        Outgoing: ChannelSpec<Priority = NoPriorityTag, Data = DataType>,
        DataType: StreamData,
    > AbstractFdService for StreamService<Incoming, Outgoing, DataType>
{
    fn ready_read(&self, api: &Fleck) {
        let mut buf = [0u8; 2048];
        while let Ok(len) = self.stream.borrow_mut().read(&mut buf) {
            api.queue::<Incoming>(Vec::from(&buf[..len]).into());
        }
    }

    fn ready_write(&self, _api: &Fleck) {
        let mut queue = self.send_queue.borrow_mut();
        while !queue.is_empty() {
            let pkt = queue.front_mut().unwrap();

            let (data, _) = queue.as_slices();

            match self.stream.borrow_mut().write(data) {
                Ok(len) => {
                    queue.drain(0..len);
                }
                Err(e) => {
                    break;
                }
            }
        }

        if queue.is_empty() {
            self.clear_write_interest()
        }
    }

    fn token(&self) -> Token {
        self.token
    }
}
