use std::any::TypeId;
use std::collections::HashMap;
use std::ops::Deref;
use std::rc::Rc;

pub(crate) mod event;

pub use event::ChannelSpec;

pub type EventRoot = event::EventRoot<crate::Fleck>;

#[allow(unused_variables)]
pub trait Service: std::any::Any + crate::helper::AsAny {
    fn register_channels(&self, eroot: &mut EventRoot) {}
    fn register_msgs(&self, registry: &mut crate::msg::MessageRegistry) {}
}

trait ServiceExt {
    fn build() -> Self;
}

impl<T: Service + Default> ServiceExt for T {
    fn build() -> Self {
        Default::default()
    }
}

#[derive(Default)]
pub struct ServiceStack {
    services: Vec<Box<dyn Service>>,

    service_map: HashMap<TypeId, Box<dyn Service>>,

    eroot: EventRoot,
    message_registry: crate::msg::MessageRegistry,
}

impl ServiceStack {
    pub fn give_service<S: 'static + Default>(&mut self)
    where
        Rc<S>: Service,
    {
        self.add_service(Rc::new(S::default()));
    }

    pub fn add_service<S: 'static>(&mut self, srv: Rc<S>)
    where
        Rc<S>: Service,
    {
        srv.register_msgs(&mut self.message_registry);
        srv.register_channels(&mut self.eroot);

        self.service_map
            .insert(TypeId::of::<S>(), Box::new(srv.clone()));
        self.services.push(Box::new(srv));
    }

    pub fn with_service<S: 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R
    where
        Rc<S>: Service,
    {
        let id = TypeId::of::<S>();

        self.service_map
            .get(&id)
            .expect("asked for service that doesn't exist!")
            .deref()
            .as_any()
            .downcast_ref::<Rc<S>>()
            .map(|s| f(s.as_ref()))
            .unwrap()
    }

    pub fn event_root(&self) -> &EventRoot {
        &self.eroot
    }

    pub fn message_registry(&self) -> &crate::msg::MessageRegistry {
        &self.message_registry
    }
}

/*
impl ServiceStack {
    pub(crate) fn process_incoming(&self, api: &crate::Fleck, packet: Packet) {
        self.eroot
            .channel::<fleck_core::ReceivePacketChannel>()
            .queue(packet);
        self.eroot.fire_all(api);
    }

    pub(crate) fn process_minor_tick(&self, api: &crate::Fleck) {
        self.eroot.channel::<fleck_core::MinorTickChannel>().queue(());
        self.eroot.fire_all(api);
    }

    pub(crate) fn process_major_tick(&self, api: &crate::Fleck) {
        self.eroot.channel::<fleck_core::MajorTickChannel>().queue(());
        self.eroot.fire_all(api);
    }

    pub(crate) fn create_io_channels(&mut self) {
        self.eroot.create_channel::<fleck_core::SendPacketChannel>();
        self.eroot.create_channel::<fleck_core::ReceivePacketChannel>();
        self.eroot.create_channel::<fleck_core::MinorTickChannel>();
        self.eroot.create_channel::<fleck_core::MajorTickChannel>();
    }
}
*/
